/*
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2011 digitalSTROM.org, Zurich, Switzerland
 *  Author: Sebastian Scholz <digitalstrom@sunshine-studios.org>
 */

Ext.define('DSS.addon.Metering', {

	extend : 'DSS.addon.Framework',

	uses : ['DSS.addon.Metering.MainPanel', 'Ext.Component'],

	config : {
		appName : _('Metering'),
		appId : 'metering',
		appVersion : '0.2.0',
		appIcon : 'images/dss/default_icon.png',
		appLang : 'de_DE'
	},

	getHelp : function() {
		return Ext.create('Ext.Component', {
			loader : {
				url : 'locale/' + this.appLang + '/LC_MESSAGES/help.html',
				autoLoad : true,
				disableCaching : false
			}
		});
	},
	getContent : function() {
		return Ext.create("DSS.addon.Metering.MainPanel", this.config);
	}
});

Ext.onReady(function() {
	var metering = Ext.create('DSS.addon.Metering');
	metering.initPage();
});

/*######################################
 static configs
 ######################################*/

DSS.addon.Metering.time = {
		2 : _('secondly'),
		60 : _('minutely'),
		300 : _('5 minutely'),
		1800 : _('half-hourly'),
		3600 : _('hourly'),
		7200 : _('2 hourly'),
		86400 : _('daily'),
};

DSS.addon.Metering.steps = {
		2 : [Ext.Date.MINUTE, 1],
		60 : [Ext.Date.MINUTE, 10],
		300 : [Ext.Date.HOUR, 1],
		1800 : [Ext.Date.DAY, 1],
		3600 : [Ext.Date.DAY, 1],
		7200 : [Ext.Date.DAY, 1],
		86400 : [Ext.Date.DAY, 1],
}

DSS.addon.Metering.refreshInterval = 2000;



/*######################################
 MainPanel
 ######################################*/
Ext.define('DSS.addon.Metering.MainPanel', {
	extend : 'Ext.panel.Panel',
	
	alias : 'widget.mainpanel',

	initComponent : function() {
		
		this.dSMTree = Ext.create('DSS.addon.Metering.dSMTree');
		this.centerPanel = Ext.create('DSS.addon.Metering.CenterPanel');
		
		Ext.apply(this, {
			//title : _('Metering Charts'),
			split : true,
			layout: {
				type: 'border'
			},
			items: [
            {
                items: this.dSMTree,
                region: 'west',
                border: true,
                split: true,
                minWidth: 50,
                width: 250
            }, 
            this.centerPanel
            ]
		});

		DSS.addon.Metering.MainPanel.superclass.initComponent.apply(this, arguments);
	}
});

/*######################################
 dSMTree
 ######################################*/
Ext.define('DSS.addon.Metering.dSMTree',
{
    extend: 'Ext.tree.Panel',
    
    
    initComponent: function()
    {
        var me = this;
        this.circuitStore = Ext.create('DSS.apartment.CircuitStore');
        this.circuitStore.on('load', function() {me.getRootNode().collapse(); me.getRootNode().expand();});
        
        Ext.apply(this, 
        {
            title: _("dSM Overview"),
            style: 'padding-bottom: 5px',
            autoScroll: true,
            animate: true,
            enableDD: false,
            containerScroll: true,
            rootVisible: true,
            store: this.circuitStore,
            border: false,
            height: 500,
            simpleSelect: false,
            multiSelect: false,
            singleSelect: true,
            listeners: {
            	select: function() {
            		this.getSelectedDSID();
            		this.findParentByType('mainpanel').centerPanel.repaint();
            		
            	}
            }
        });

        DSS.addon.Metering.dSMTree.superclass.initComponent.apply(this, arguments);
    },
    
    getSelectedDSID: function()
    {
        if(this.getSelectionModel().getCount()==0){
        	this.getSelectionModel().select(0);
        }
                
        var node = this.getSelectionModel().getSelection();
        if (node !== null) {          
            var dsid = node[0].get('dsid');
            if(dsid === 0) {
            	//whole apartment is selected -> select next dsm
            	this.getSelectionModel().select(1);
            	node = this.getSelectionModel().getSelection();
            	if (node !== null) {
            		dsid = node[0].get('dsid');
            	}
            }
            return dsid;
        }
        else 
            Ext.MessageBox.alert(_('Error'), _('please select a dSM!'));
    }
});

/*######################################
 CenterPanel
 ######################################*/

Ext.define('DSS.addon.Metering.CenterPanel', {
	extend : 'Ext.panel.Panel',

	initComponent : function() {
		var me = this;

		this.unit = '';

		this.resolutionStore = Ext.create('DSS.apartment.ResolutionStore');
		this.resolutionStore.load();

		if(this.resolutionStore.isLoading()) {
			this.resolutionStore.on('load', function() {
				me.initStores();
			});
		} else {
			this.initStores();
		}

		this.consumptionStore = Ext.create('Ext.data.Store', {
			model : 'DSS.apartment.ResolutionModel'
		});

		this.energyStore = Ext.create('Ext.data.Store', {
			model : 'DSS.apartment.ResolutionModel'
		});

		this.typeStore = Ext.create('Ext.data.Store', {
			fields : ['type', 'name'],
			data : [{
				"type" : "consumption",
				"name" : _('Consumption')
			}, {
				"type" : "energy",
				"name" : _('Energy')
			}]
		});

		this.typeBox = Ext.create('Ext.form.field.ComboBox', {
			store : this.typeStore,
			labelAlign : 'top',
			editable : false,
			displayField : 'name',
			valueField : 'type',
			fieldLabel : _('unit')
		});
		this.typeBox.on('change', function() {
			me.repaint();
		});

		this.energyTimeBox = Ext.create('Ext.form.field.ComboBox', {
			store : this.energyStore,
			labelAlign : 'top',
			editable : false,
			queryMode : 'local',
			displayField : 'name',
			valueField : 'resolution',
			fieldLabel : _('resolution')
		});
		this.energyTimeBox.on('change', function() {
			me.repaint();
		});

		this.consumptionTimeBox = Ext.create('Ext.form.field.ComboBox', {
			store : this.consumptionStore,
			labelAlign : 'top',
			editable : false,
			queryMode : 'local',
			displayField : 'name',
			valueField : 'resolution',
			fieldLabel : _('resolution')
		});
		this.consumptionTimeBox.on('change', function() {
			me.repaint();
		});
		this.consumptionTimeBox.setVisible(false);

		this.meteringStore = Ext.create('Ext.data.Store', {
			model : 'DSS.addon.Metering.MeteringModel',
			proxy : {
				type : 'ajax',
				actionMethods : {
					read : 'GET'
				},
				pageParam : undefined,
				limitParam : undefined,
				startParam : undefined,
				reader : {
					type : 'array',
					root : 'result.values',
					successProperty : 'ok'
				}
			},
			sorters : [{
				property : 'date',
				direction : 'ASC'
			}],
			// override remove method of this store so that no 'datachanged' or 'remove' event is fired
			// -> chart is only redrawed when a new value pair is added to the store
			// -> better performance
			remove: function(records, /* private */ isMove) {
				if(!Ext.isArray(records)) {
					records = [records];
				}
				/*
				 * Pass the isMove parameter if we know we're going to be re-inserting this record
				 */
				isMove = isMove === true;
				var me = this, sync = false, i = 0, length = records.length, isPhantom, index, record;

				for(; i < length; i++) {
					record = records[i];
					index = me.data.indexOf(record);

					if(me.snapshot) {
						me.snapshot.remove(record);
					}
					if(index > -1) {
						isPhantom = record.phantom === true;
						if(!isMove && !isPhantom) {
							// don't push phantom records onto removed
							me.removed.push(record);
						}
						record.unjoin(me);
						me.data.remove(record);
						sync = sync || !isPhantom;
					}
				}

				if(!isMove && me.autoSync && sync) {
					me.sync();
				}
			}
		});

		this.chart = Ext.create('Ext.chart.Chart', {
			mask : 'horizontal',
			store : this.meteringStore,
			axes : [{
				title : _('Value'),
				type : 'Numeric',
				position : 'left',
				fields : ['value'],
				minimum : 0/*,
				grid : {
					odd : {
						opacity : 1,
						fill : '#d6f5e4'
					},
					even : {
						opacity : 1,
						fill : '#baefd1'
					}
				}*/
			}, {
				type : 'Time',
				position : 'bottom',
				fields : 'date',
				title : _('Time'),
				dateFormat : 'd. M - G:i',
				groupBy : 'hour,minute,second',
				aggregateOp : 'sum',
				//constrain: true
				//fromDate: new Date(2011, 10, 4),
				//toDate: new Date(2011, 10, 6)
			}],
			series : [{
				type : 'line',
				xField : 'date',
				yField : 'value',
				axis : ['left', 'bottom'],
				smooth : false,
				showMarkers : false,
				style: {
                    fill: '#00523b',
                    stroke: '#00523b',
                    'stroke-width': 3
               },
				tips : {
					trackMouse : true,
					width : 140,
					height : 35,
					renderer : function(storeItem, item) {
						this.setTitle(storeItem.get('value') + me.unit + ' ' + _('at') + ' ' + Ext.Date.format(new Date(storeItem.get('date')), 'd.m.Y - G:i:s'));
					}
				},
				jumpMarkers: [],
				addJumpMarker : function(storeItem, text) {
					var meToo = this;

					for(var i = 0, ln = meToo.items.length; i<ln; i++){		
						if(storeItem.get('date') == meToo.items[i].value[0] && storeItem.get('value') == meToo.items[i].value[1]){
							console.log('match');
							
							var textSprite = Ext.create('Ext.draw.Sprite', {
								type : 'text',
								surface : me.chart.surface,
								text : text,
								x : meToo.items[i].point[0] - 40,
								y : meToo.items[i].point[1] - 20,
								storeItem: storeItem
							});
							
							/*var frameSprite = Ext.create('Ext.draw.Sprite', {
								type : 'path',
								surface : me.chart.surface,
								path : "M 0 10 L 0 40 Q 0 50 10 50 L 50 50 L 60 70 L 70 50 L 110 50 Q 120 50 120 40 L 120 10 Q 120 0 110 0 L 10 0 Q 0 0 0 10 Z",
								x : meToo.items[i].point[0],
								y : meToo.items[i].point[1],
								stroke : '#d6f5e4',
								fill :  '#baefd1',
								storeItem : storeItem
							});
									
							frameSprite.setAttributes({
								translate : {
									x : meToo.items[i].point[0]-60,
									y : meToo.items[i].point[1]-90
								}
							}, true);*/

							textSprite.show(true);
							//frameSprite.show(true);
							meToo.jumpMarkers.push(textSprite);
							//meToo.jumpMarkers.push(frameSprite);					
						}
					}
				},
				getPositionForStoreItem: function(storeItem) {
					var meToo = this;

					for(var i = 0, ln = meToo.items.length; i<ln; i++){
						if(storeItem.get('date') == meToo.items[i].value[0] && storeItem.get('value') == meToo.items[i].value[1]){	
							return meToo.items[i].point
						}
					}
				},
				updateMarkers: function() {
					var meToo = this;
					
					for(var i = 0; i<meToo.jumpMarkers.length; i++){
						var sprite = meToo.jumpMarkers[i];
						var point = meToo.getPositionForStoreItem(sprite.storeItem);
						if(point === undefined){
							meToo.jumpMarkers.splice(i,1);
							sprite.remove();
						} else {
							sprite.setAttributes({
								x : point[0] - 40,
								y : point[1] - 20
							}, true);
						}
					}
				}
			}],
			theme : 'MeteringTheme',
			listeners : {
				select : {
					fn : function(me, selection) {
						me.setZoom(selection);
						me.mask.hide();
					}
				}
			}
		});

		this.chart.on('refresh', function() {me.chart.series.get(0).updateMarkers();})

		this.zoomOutButton = Ext.create(Ext.button.Button, {
			text : _('reset zoom'),
			handler : function() {
				me.chart.restoreZoom();
			}
		});
		
		this.thread = null;

		Ext.apply(this, {
			border : false,
			layout : 'fit',
			items : this.chart,
			tbar : [this.typeBox, this.energyTimeBox, this.consumptionTimeBox, this.zoomOutButton],
			region : 'center',
			border : true,
			split : true,
			minWidth : 600
		});

		DSS.addon.Metering.CenterPanel.superclass.initComponent.apply(this, arguments);
	},
	initStores : function() {
		for( i = 0; i < this.resolutionStore.getCount(); i++) {
			if(this.resolutionStore.getAt(i).get('type') === 'consumption') {
				var record = this.resolutionStore.getAt(i);
				record.setName(DSS.addon.Metering.time[record.get('resolution')]);
				this.consumptionStore.add(record);
			} else {
				var record = this.resolutionStore.getAt(i);
				record.setName(DSS.addon.Metering.time[record.get('resolution')]);
				this.energyStore.add(record);
			}
		}
				
		this.typeBox.setValue('consumption');
		//this.energyTimeBox.setValue(this.energyStore.getAt(0).get('resolution'));
		//this.consumptionTimeBox.setValue(this.consumptionStore.getAt(0).get('resolution'));
		this.unit = this.consumptionStore.getAt(0).get('unit');
	},
	repaint : function() {
		var dsid = this.findParentByType('mainpanel').dSMTree.getSelectedDSID();
		var url;
		
		if(this.typeBox.getValue() === 'consumption') {
			this.consumptionTimeBox.setVisible(true);
			this.energyTimeBox.setVisible(false);
			this.unit = this.consumptionStore.getAt(0).get('unit');
			this.chart.axes.get(1).step = DSS.addon.Metering.steps[this.consumptionTimeBox.getValue()];
			url = '/json/metering/getValues?dsid=' + dsid + '&type=consumption&resolution=' + this.consumptionTimeBox.getValue();
		} else {
			this.consumptionTimeBox.setVisible(false);
			this.energyTimeBox.setVisible(true);
			this.unit = this.energyStore.getAt(0).get('unit');
			this.chart.axes.get(1).step = DSS.addon.Metering.steps[this.energyTimeBox.getValue()];
			url = '/json/metering/getValues?dsid=' + dsid + '&type=energy&resolution=' + this.energyTimeBox.getValue();
		}

		this.meteringStore.removeAll();
		this.meteringStore.getProxy().url = url;
		this.meteringStore.load();

		this.chart.axes.get(0).setTitle(this.typeBox.getRawValue() + ' (' + this.unit + ')');

		this.chart.restoreZoom();

		if(this.typeBox.getValue() === 'consumption' && this.consumptionTimeBox.getValue() === 2 && this.thread === null) {
			this.startThread();
		} else {
			this.stopThread();
		}
	},
	startThread : function() {
		var me = this;
		
		this.thread = setInterval(function() {
			Ext.Ajax.request({
				url : '/json/circuit/getConsumption?id=' + me.findParentByType('mainpanel').dSMTree.getSelectedDSID(),
				success : function(response) {
					var jsonData = Ext.JSON.decode(response.responseText);
					if(jsonData.ok) {
						var array = [];
						var record = [Math.round(+new Date() / 1000), jsonData.result.consumption]
						array.push(record);

						me.meteringStore.removeAt(0);
						me.meteringStore.loadData(array, true);

						/*   detect jumps
						* jump detection is based on the last 3 values
						* 
						*              c----d
						*             /
						*            /
						*           x 
						*          /
						*         /
						*   a----b
						* 
						* a jump is detected for c, if
						*      |a-b| < 5% of a
						*      |a-c| > 25% of a
						*      |c-d| < 5% of c
						*      |c-x| > 15% of c
						* 
						*/
						
						
						var count = me.meteringStore.getCount();
						var a = me.meteringStore.getAt(count - 5).get('value');
						var b = me.meteringStore.getAt(count - 4).get('value');
						var x = me.meteringStore.getAt(count - 3).get('value');
						var c = me.meteringStore.getAt(count - 2).get('value');
						var d = me.meteringStore.getAt(count - 1).get('value');
						
						if(Math.abs(a-b)<=0.05*a && Math.abs(c-d)<=0.05*c && Math.abs(a-c)>0.25*a && Math.abs(c-x)>0.15*c) {
							var diff = c - a;
							var diffString;
							if(diff>0){
								diffString = '+'+diff+' W';
							} else {
								diffString = diff+' W';
							}							
							me.chart.series.get(0).addJumpMarker(me.meteringStore.getAt(me.meteringStore.getCount()-2), diffString);
						}
					}
				}
			});
		}, DSS.addon.Metering.refreshInterval);
	},
	stopThread : function() {
		clearInterval(this.thread);
	}
});



/*######################################
 MeteringModel
######################################*/

Ext.define('DSS.addon.Metering.MeteringModel', {
	extend : 'Ext.data.Model',
	fields : [
		{name : 'date', type: 'date', dateFormat: 'timestamp', mapping: 0},
		{name : 'value', mapping: 1}
	]
});


/*######################################
 MeteringTheme
######################################*/


Ext.define('Ext.chart.theme.MeteringTheme', {
	extend : 'Ext.chart.theme.Base',

	constructor : function(config) {
		var baseColor = '#005138';
		//light green (like the borders)
		//var background = '#99e8bc';
		//darker green (like dSM Overview)
		//var fill = '#048c40';
		//very dark green (like the dS Logo)
		var fill = '#00523b';

		this.callParent([Ext.apply({
			axis : {
				fill : baseColor,
				stroke : baseColor
			},
			axisLabelLeft : {
				fill : baseColor
			},
			axisLabelBottom : {
				fill : baseColor
			},
			axisTitleLeft : {
				fill : baseColor
			},
			axisTitleBottom : {
				fill : baseColor
			},
			colors : [fill],
			markerThemes : [{
				fill : baseColor,
				type : 'circle'
			}]
		}, config)]);
	}
});


